extends Node
# スプラッシュ画面から最初に遷移するシーン専用のスクリプトです
# このスクリプトがアタッチされたシーンを土台として、ゲームの基本機能を使用したり、インスタンスを管理します
# 含まれる機能: インスタンスアセットのプリロードと呼び出し、サウンドの再生管理、セーブファイルの管理
@onready var bgm: AudioStreamPlayer = $BGM
@onready var bgm_mirror: AudioStreamPlayer = $BGMMirror
@onready var se: AudioStreamPlayer = $SE

# プリロードによって自動生成されたアセット変数を格納するための辞書
var preload_dict = {}
# 対象ファイルのプリロード全完了判定
var preload_complete: bool = true

func _ready():
	# Data.gdを中継としたシグナルによって、各機能を他のスクリプトから呼び出せるようにする
	Data.connect("node", Callable(self, "_node"))
	Data.connect("set_image", Callable(self, "_set_image"))
	Data.connect("audio", Callable(self, "_audio"))
	Data.connect("manage_save_data", Callable(self, "_manage_save_data"))
	Data.connect("get_preloaded_instance", Callable(self, "_get_preloaded_instance"))
	Data.connect("has_temporary_slot_data", Callable(self, "_has_temporary_slot_data"))
	Data.connect("update_file", Callable(self, "_update_file"))
	Data.connect("restart_for_transfer", Callable(self, "_ready"))
	# スロットごとの一時セーブデータのセットを試みる
	if _has_temporary_slot_data(Data.data_templates_directory, Data.per_slot_data_file_name, true):
		# awaitでインスタンスアセットのプリロードを完了後、アプリを開始する
		await preload_all_instances()
		# 全てのプリロード処理に問題がなければスタート
		if preload_complete:
			_node("add", "Start")
		else:
			push_error("Preload Failed.", self)

# インスタンスシーン（ノード）の生成と削除
func _node(command: String, node_name):
	# PackedScene型のオブジェクトを、実際のシーンやノードのインスタンスに変換する
	var scene_instance = _get_preloaded_instance(node_name).instantiate()
	# インスタンスがプリロードされている場合
	if scene_instance != null:
		match command:
			"add": # ノードを生成して追加
					add_child(scene_instance)
			"delete": # 親ノードから削除（内部情報は保持）
				if has_node(scene_instance):
					remove_child(scene_instance)
			"free": # 親ノードから削除してメモリを解放
				if has_node(scene_instance):
					scene_instance.queue_free()
	else:
		push_error("Signal Error node: " + scene_instance + "インスタンスが見つかりません。", self)

# StoryLib専用: 指定したノードにテクスチャファイルをセットする
func _set_image(main_node, under_node, texture_name):
	var preloaded_texture = _get_preloaded_instance(texture_name)
	# 表示可能な形式か確認する
	if preloaded_texture != null and preloaded_texture is Texture:
		main_node.set_texture(preloaded_texture)
		# under_nodeが指定されている場合
		if under_node != null:
			under_node.set_texture(null)
	else:
		push_error("Error: 表示可能な形式の" + texture_name + "テクスチャが見つかりません。", self)

# サウンドコントロール（再生、フェードイン、フェードアウト、停止）
func _audio(command: String, title, is_bgm: bool):
	var sound
	# BGMかSEか
	if is_bgm:
		sound = bgm
	else:
		sound = se
	
	match command:
		"play":
			if title != null:
				if is_bgm: # BGMの場合
					bgm_mirror.set_stream(_get_preloaded_instance(title))
					
					# BGM停止中または同じ曲でなければ最初から再生する
					if (bgm.get_stream() != bgm_mirror.get_stream()) or !bgm.playing:
						bgm.set_volume_db(0)
						bgm.set_stream(_get_preloaded_instance(title))
						bgm.playing = true
				else: # true以外はSE扱い
						se.set_stream(_get_preloaded_instance(title))
						se.playing = true
		"stop":
			sound.playing = false
		_:
			# 指定ノードに音声ファイルが格納されている場合
			if sound.get_stream() != null:
				match command:
					"fadein":
						# ボリュームを調整してから再生再開
						sound.set_volume_db(-30)
						sound.stream_paused = false
						
						for i in range(-30, 1, 5):  # -30から0まで0.25秒ごとに5ずつボリュームを上げていく
							sound.set_volume_db(i)
							await get_tree().create_timer(0.25).timeout
						# ボリュームを最大に設定
						sound.set_volume_db(0)
					"fadeout":
						for i in range(-5, -61, -5): # -5から-60まで0.25秒ごとに-5ずつボリュームを下げていく
							sound.set_volume_db(i)
							await get_tree().create_timer(0.25).timeout
						# 一時停止後にボリュームを戻す
						sound.stream_paused = true
						sound.set_volume_db(0)
					"pause":
						sound.stream_paused = true
					"resume":
						sound.stream_paused = false
					_:
						push_error("Error: commandに適切な文字列を使用してください。", self)

# セーブファイルの管理
func _manage_save_data(command: String, file_name: String, extension_name: String, is_encrypted: bool):
	var directory_path = ""
	var keyword = command.split(" ", true, 1)
	# 拡張子によって参照ディレクトリを振り分ける
	match extension_name:
		"json":
			directory_path = Data.root_directory
		_:
			push_error("Error manage_save_data: 対応している拡張子を指定してください。", self)
			return
	
	# 指定キーの要素を更新する 例: update second_nest>second_bool, true
	if command.begins_with("update"): # 個別項目の更新
		var key_data = keyword[1].split(", ", true, 1) # コンマスペースごとにコマンドを2つに分割
		var directory_name = key_data[0].split(">", true, -1) # 階層判定用（-1で無制限に分割）
		var reference_data = Data.current_temporary_slot_data
		var is_integer = true
		var optimized_data = key_data[1] # 未処理のままであればString型として更新される
		# 階層チェック
		for i in range(directory_name.size() - 1):
			# 指定の階層が存在する場合
			if reference_data.has(directory_name[i]):
				reference_data = reference_data[directory_name[i]]
			else:
				push_error("Error manage_save_data: 指定した要素が見つかりませんでした。", self)
				return
		# 数字の文字セットを定義
		var numeric_chars = ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"]
		# 更新データを一文字ずつ確認する
		for char in key_data[1]:
			# 数字でなければ処理を抜ける
			if not char in numeric_chars:
				is_integer = false
				break
		# 数字であればintに変換する
		if is_integer:
			optimized_data = int(key_data[1])
		else: # 更新データがbool型を想定していた場合
			# 全て小文字に変換する
			var check_data = key_data[1].to_lower()
			# bool型として代入する
			if check_data == "true":
				optimized_data = true
			elif check_data == "false":
				optimized_data = false
		# 一時データの指定のキーを更新する
		reference_data[directory_name[directory_name.size() - 1]] = optimized_data
		Data.current_temporary_slot_data = reference_data
		return
	
	var file_path = directory_path + "/" + file_name + "." + extension_name
	# 指定のローカルディレクトリが見つかった場合
	if DirAccess.dir_exists_absolute(directory_path):
		# 指定スロットをセーブする 例: save_per_slot 1
		if command.begins_with("save_per_slot"):
			# セーブ可能なスロット数の範囲内に収まっている場合
			if int(keyword[1]) <= Data.max_save_slot:
				# 指定のキーが現在ロード中のセーブスロットと同じ場合
				if keyword[1] == Data.current_save_slot:
					# 一時データの中に指定のキーが存在する場合、全てのキーの照合を確認後、一時データを更新する
					if keyword[1] in Data.temporary_saves:
						for key in Data.current_temporary_slot_data:
							if key in Data.temporary_saves[keyword[1]]:
								Data.temporary_saves[keyword[1]][key] = Data.current_temporary_slot_data[key]
					else: # ない場合は指定スロットを新規作成する
						Data.temporary_saves[keyword[1]] = Data.current_temporary_slot_data
				# スロットが異なる場合は指定スロットを新規作成する
				else:
					Data.temporary_saves[keyword[1]] = Data.current_temporary_slot_data
				# 一時データの内容をローカルディレクトリに保存する
				_update_file(directory_path, file_name, extension_name, Data.temporary_saves, is_encrypted)
			else:
				push_error("Error manage_save_data: セーブ可能なスロット数より大きな値が指定されています。", self)
		# 指定ファイルが見つかった場合
		elif FileAccess.file_exists(file_path):
			# 指定スロットをロードする 例: load_per_slot 2
			if command.begins_with("load_per_slot"):
				# 一時データの中に指定のキーが存在する場合、指定キーの中身をロードする
				if keyword[1] in Data.temporary_saves:
					Data.current_temporary_slot_data = Data.temporary_saves[keyword[1]]
					Data.current_save_slot = keyword[1]
				else:
					push_error("Error manage_save_data: 指定のロード用セーブデータが見つかりません。", self)
			# 一時データの中に指定のキーが存在する場合、指定キーを削除する 例: delete_per_slot 3
			elif command.begins_with("delete_per_slot"):
				if keyword[1] in Data.temporary_saves:
					Data.temporary_saves.erase(keyword[1])
					# セーブデータが全て空の場合、セーブファイルを削除する
					if Data.temporary_saves == {}:
						DirAccess.remove_absolute(file_path)
					else:
						# 一時データの内容をローカルディレクトリに保存する
						_update_file(directory_path, file_name, extension_name, Data.temporary_saves, is_encrypted)
			# データをファイルごと消す
			elif command.begins_with("delete_all"):
				DirAccess.remove_absolute(file_path)
			else:
				push_error("Error manage_save_data: 有効なコマンドが指定されていません。", self)
		else:
			push_error("Error manage_save_data: 指定ファイルが見つかりません。", self)
	else:
		push_error("Error manage_save_data: 指定ディレクトリが見つかりません。", self)

# 自動生成された変数を、文字列を使って呼び出す関数
func _get_preloaded_instance(var_name: String) -> Object:
	var_name = var_name.strip_edges() # 文字列を正常に認識させるために余白を削除
	
	if var_name in preload_dict:
		return preload_dict[var_name]
	else:
		return null

# スロットごとの一時セーブデータのセットを試みて結果を返す
func _has_temporary_slot_data(temp_file_path, temp_file_name, do_init) -> bool:
	var initial_local_save_path = temp_file_path + temp_file_name + ".json"
	var user_local_save_path = Data.root_directory + Data.save_file_name + "." + Data.save_file_extension
	# セーブデータのテンプレ―トファイルが見つかった場合
	if FileAccess.file_exists(initial_local_save_path):
		# 一時セーブデータを初期化する
		if do_init:
			# 複合化処理なしで内部ディレクトリからテンプレートデータを取得する
			var per_slot_data = Data.get_file_text(initial_local_save_path, false)
			# 初期データを一時データに代入する
			Data.current_save_slot = "0"
			Data.current_temporary_slot_data = per_slot_data
			# ユーザーのローカルフォルダにセーブデータが存在していればそれを取得する
			if FileAccess.file_exists(user_local_save_path):
				Data.temporary_saves = Data.get_file_text(user_local_save_path, Data.use_encrypt)
				# per_slot_dataとData.temporary_savesのそれぞれのセーブスロットのデータを確認して、後者のキーに抜けがあれば前者のキーを追加する
		return true
	else:
		push_error("Error manage_save_data: 指定のテンプレートデータが見つかりません。", self)
		return false

# ファイルの新規作成と更新
func _update_file(directory_path, file_name, extention_name, update_data, is_encrypted):
	var directory = directory_path + file_name + "." + extention_name
	var file_access
	# 暗号化判定
	if is_encrypted:
		file_access = FileAccess.open_encrypted_with_pass(directory, FileAccess.WRITE, Data.encrypt_key)
	else:
		file_access = FileAccess.open(directory, FileAccess.WRITE)
	# ファイルが存在している場合、内容をupdate_dataに更新する
	if file_access != null:
		file_access.store_line(JSON.new().stringify(update_data))
	
	file_access.close()

# Preloadsフォルダ以下にある全てのファイルを再帰的にpreloadし、変数を生成する
func preload_all_instances(preload_path: String = "res://Instances"):
	var dir = DirAccess.open(preload_path)
	# ディレクトリが存在している場合
	if dir and dir.change_dir(preload_path) == OK:
		dir.list_dir_begin()
		var file = dir.get_next()
		var file_count = 0
		# ファイルを全て確認するまで繰り返す
		while file != "":
			var file_path = preload_path + "/" + file
			# ファイルの拡張子をチェック
			if file.get_extension().to_lower() in ["ogg", "png", "tscn"]:
				# ファイル名に日本語や特殊文字が含まれているかチェック
				for i in range(file.length()):
					if file.unicode_at(i) > 127:
						preload_complete = false
						push_error("Error preload_all_instances: ファイル名に16ビット文字や特殊文字が含まれています。ファイル名:", file)
						return
				# preloadしたファイルを扱うための変数を自動生成する
				var var_name = file.get_basename().replace(".", "_")
				# すでにプリロードされた変数と名前が被る場合、処理を中止する
				for key in preload_dict.keys():
					if key == var_name:
						preload_complete = false
						push_error("Error preload_all_instances: 名前が重複しているプリロードアセットがあります。ファイル名:", file)
						return
				# ファイルをResourceLoaderを使って読み込む(awaitで読み込みが完了するまで待機する)
				var preload_file = await ResourceLoader.load(file_path)
				# 読み込んだファイルがある場合、変数を辞書に格納し、呼び出し可能にする
				if preload_file != null:
					set(var_name, preload_file)
					preload_dict[var_name] = preload_file
					file_count += 1
			
			file = dir.get_next()
		
		if file_count > 0:
			print(preload_path, "から", file_count, "のアセットをプリロードしました。")
		
		dir.list_dir_end()
		dir.list_dir_begin()
		file = dir.get_next()
		# サブフォルダがあれば再帰的に検索
		while file != "":
			if dir.current_is_dir():
				preload_all_instances(preload_path + "/" + file)
			
			file = dir.get_next()
	else:
		push_error("Error: 指定されたフォルダが見つかりませんでした:", preload_path)

# ランダム整数の取得
func get_random(int_min, int_max): 
	return randi_range(int_min, int_max + 1)
